module BWT where

import List
import Data.Array

----------------------------
-- Hilfsfunktionen
----------------------------

{-
Position
--------
Gibt die Position eines Listen-Elementes aus

Beispielaufruf:
position "cab" ["abc","bca","cab"] = 2
-}
position :: Eq a => a -> [a] -> Int
position xs xss = length ( takeWhile (/=xs) xss )

{-
lrot
--------
Rotiert eine Liste um einen nach links

Beispielaufruf:
lrot "cab" = "abc"
-}
lrot :: Ord a => [a] -> [a]
lrot (x:xs) = xs ++ [x]

{-
rrot
--------
Rotiert eine Liste um einen nach rechts

Beispielaufruf:
rrot "cab" = "bca"
-}
rrot :: Ord a => [a] -> [a]
rrot [] = []
rrot xs = [last xs] ++ init xs

{-
ROTate Sequence
---------------
Schreibt jede Rotation der Eingabeliste in die Resultat-Map

Beispielaufruf:
rots "cab" =
["cab",
 "abc",
 "bca"]
-}
rots :: Ord a => [a] -> [[a]]
rots xs = take (length xs) (iterate lrot xs)

{-
hdsort
--------
Sortiert die erste Spalte einer Matrix

Beispielaufruf:
hdsort ["bc","ca","ab"] = ["ab","bc","ca"]
-}
hdsort :: Ord a => [[a]] -> [[a]]
hdsort = sortBy cmp 
         where
         cmp (x:xs) (y:ys) = compare x y
		 
{-
consCol
--------
Fgt eine neue Spalte zu einer Matrix hinzu

Beispielaufruf:
consCol ( "abc", ["bc","ca","ab"]) = ["abc","bca","cab"]
-}
consCol :: ([a],[[a]]) -> [[a]]
consCol (xs,xss) = zipWith (:) xs xss

{-
takeCols
--------
Auslesen der ersten j Spalten einer Matrix
takeCols n = id

Beispielaufruf:
takeCols 2 ["abc","bca","cab"] = ["ab","bc","ca"]
-}
takeCols :: Int -> [[a]] -> [[a]]
takeCols j = map(take j)

{-
recreate
--------
Berechnet die n x n Matrix Spalte fr Spalte

Beispielaufruf:
recreate (length "cab") "cab" = ["abc","bca","cab"]
-}
recreate :: Ord a => Int -> [a] -> [[a]]
recreate 0     = map (const [])
recreate (j+1) = hdsort.consCol.fork (id,recreate j)
                 where
                 fork (f,g) x = (f x,g x)
				 
---------------
-- PIMPUP #1 --
---------------

apply :: (Ord a) => [Int] -> [a] -> [a]
apply px xs   = [ xs !!(px !! i) | i <- [0..(length(xs)-1)]]


permutation :: (Ord a) => [a] -> [Int]
permutation xs = map snd (sort (zip xs [0..]))


{-
recreate2
--------
Berechnet die n x n Matrix Spalte fr Spalte

hdsort, welches die erste Spalte einer Matrix sortiert,
muss ersetzt werden

Beispielaufruf:
recreate (length "cab") "cab" = ["abc","bca","cab"]
-}
recreate2 :: Ord a => Int -> [Int] -> [a] -> [[a]]
recreate2 0 _ xs    = map (const []) xs
recreate2 (j+1) p xs = consCol.fork(apply p, apply p . recreate2 j p) $ xs
                 where
                 fork (f,g) x = (f x,g x)

-- PIMPUP #2

recreate3 :: Ord a => Int -> [a] -> [[a]]
recreate3 j xs = transpose.(take j).tail.iterate (apply p) $ xs
                 where
                 p = permutation xs

--------------------------
-- Hauptfunktionen
----------------------------

{-
Transformation
--------------
Transformiert eine Liste und gibt ein Tupel zurck,
bestehend aus der transformierten Liste und der Position dieser Liste
in der Transformationsmatrix

Beispielaufruf:
transform "abc" = ("cab",2)
-}
transform :: Ord a => [a] -> ([a],Int)
transform [] = ([],0)
transform xs = (map last xss, position xs xss)
                where xss = sort (rots xs)

-- Optimieren der Transformation durch das Nutzen von Arrays (Indizierter Zugriff) 
-- und das Nutzen eines eof-Zeichens (eof muss in der Ordnung kleiner sein als jedes 
-- andere Zeichen in der Liste xs)
transform2 :: Ord a => a -> [a] -> ([a],Int)
transform2 eof xs = ([ xa ! (pa ! i) | i <- [0..n-1]],k)
                where
                n = length xs
                k = length (takeWhile ((/=)0) ps)
                xa = listArray (0,n-1) (rrot xs)
                pa = listArray (0,n-1) ps
                ps = map snd (sort (zip (tails (tag xs))[0..(n-1)]))
                tag xs = xs ++ [eof]
{-
Rcktransformation
------------------
Bildet aus einem Tupel,bestehend aus der transformierten 
Liste und der Position dieser Liste in der Transformationsmatrix, 
die komplette, als Basis fr die Transformationsmatrix dienende, 
Liste wieder her

Beispielaufruf:
untransform ("cab",2) = "abc"
-}
untransform :: Ord a => ([a],Int) -> [a]
untransform (ys,k) = (recreate n ys) !! k
                     where n = length ys

-- Optimierung der Rcktransformation 
-- durch das Ausnutzen der immer gleichen Sortier-Reihenfolge (siehe p)
untransform2 :: Ord a => ([a],Int) -> [a]
untransform2 (ys,k) = (recreate2 n p ys) !! k
                     where 
                     n = length ys
                     p = permutation ys

-- Optimierung der Rcktransformation durch Auflsen der Rekursion des recreates			 
untransform3 :: Ord a => ([a],Int) -> [a]
untransform3 (ys,k) = (recreate3 n ys) !! k
                     where n = length ys
					 
-- Optimierung der Rcktransformation durch das Ausnutzen 
-- der konstanten Zugriffszeit auf Elemente eines Arrays	
untransform4 :: Ord a => ([a],Int) -> [a]
untransform4 (ys,k) = ((take n) .tail .map (ya!)) (iterate (pa!) k)
                      where
                      n  = length ys
                      ya = listArray (0,n-1) ys
                      pa = listArray (0,n-1) (map snd (sort (zip ys [0..])))